@@ -1,7 +1,7 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/misc.c 2.21 1998/06/20 15:13:49 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/misc.c 2.32 1999/04/20 17:41:38 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.2a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4c.
   Miscellaneous HTTP / HTML functions.
   ******************/ /******************
   Written by Andrew M. Bishop
@@ -27,22 +27,7 @@
 #include "misc.h"
 #include "errors.h"
 #include "proto.h"
-
-/* Allow this file to be compiled without needing to use the configuration
-   file reading functions in config.c.
-   - Requires a replacement function.  */
-
-#ifdef NO_CONFIG
-
-#define GetLocalHost(x) "localhost:8080"
-#define IsLocalHost(x,y) 0
-
-#else
-
 #include "config.h"
-
-#endif /* NO_CONFIG */
-
 #include "md5.h"
 
 
@@ -114,7 +99,7 @@ URL *SplitURL(char *url)
 
  /* Password */
 
- if(at && (!slash || slash>at))
+ if(at && at>copyurl && (!slash || slash>at))
    {
     if(colon && at>colon)               /* user:pass@www.foo.com... */
       {
@@ -135,6 +120,9 @@ URL *SplitURL(char *url)
    }
  else
    {
+    if(at==copyurl)             /* @www.foo.com... */
+       copyurl++;
+
     Url->user=NULL;
     Url->pass=NULL;
    }
@@ -225,6 +213,29 @@ URL *SplitURL(char *url)
     strcpy(Url->name+n+1,Url->args);
    }
 
+ if(Url->user)
+   {
+    Url->file=(char*)malloc(strlen(Url->proto)+strlen(Url->hostp)+strlen(Url->user)+(Url->pass?strlen(Url->pass):0)+8);
+
+    n=Url->hostp-Url->name;
+    strncpy(Url->file,Url->name,n);
+
+    strcpy(Url->file+n,Url->user);
+    n+=strlen(Url->user);
+
+    if(Url->pass)
+      {
+       strcpy(Url->file+n,":");
+       strcpy(Url->file+n+1,Url->pass);
+       n+=strlen(Url->pass)+1;
+      }
+
+    strcpy(Url->file+n,"@");
+    strcpy(Url->file+n+1,Url->hostp);
+   }
+ else
+    Url->file=Url->name;
+
  if(Url->Protocol && !Url->Protocol->proxyable)
    {
     char *localhost=GetLocalHost(1);
@@ -242,6 +253,49 @@ URL *SplitURL(char *url)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Add a password to an existing URL.
+
+  URL *Url The URL to add the username and password to.
+
+  char *user The username.
+
+  char *pass The password.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+void AddURLPassword(URL *Url,char *user,char *pass)
+{
+ int n;
+
+ Url->user=(char*)malloc(strlen(user)+1);
+ strcpy(Url->user,user);
+
+ if(pass)
+   {
+    Url->pass=(char*)malloc(strlen(pass)+1);
+    strcpy(Url->pass,pass);
+   }
+
+ Url->file=(char*)malloc(strlen(Url->name)+strlen(Url->user)+(Url->pass?strlen(Url->pass):0)+8);
+
+ n=Url->hostp-Url->name;
+ strncpy(Url->file,Url->name,n);
+
+ strcpy(Url->file+n,Url->user);
+ n+=strlen(Url->user);
+
+ if(Url->pass)
+   {
+    strcpy(Url->file+n,":");
+    strcpy(Url->file+n+1,Url->pass);
+    n+=strlen(Url->pass)+1;
+   }
+
+ strcpy(Url->file+n,"@");
+ strcpy(Url->file+n+1,Url->hostp);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Free the memory in a URL.
 
@@ -253,6 +307,9 @@ void FreeURL(URL *Url)
  if(Url->name!=Url->link)
     free(Url->link);
 
+ if(Url->name!=Url->file)
+    free(Url->file);
+
  free(Url->name);
 
  if(Url->proto) free(Url->proto);
@@ -285,6 +342,11 @@ char *LinkURL(URL *Url,char *link)
 
  if(colon && slash && colon<slash)
     ;
+ else if(!*link)
+   {
+    new=(char*)malloc(strlen(Url->name));
+    strcpy(link,Url->name);
+   }
  else
    {
     if(*link=='/')
@@ -431,10 +493,48 @@ char *URLEncode(const char *str)
  return(copy);
 }
 
-/* Allow this file to be compiled so that the URL manipulation functions
-   only are available and not the hashing ones. */
 
-#ifndef URL_ONLY
+/*++++++++++++++++++++++++++++++++++++++
+  Encode a string using the UrlEncode method as used for passwords in URLs.
+
+  char *UrlEncodePassword Returns a malloced copy of the encoded string.
+
+  const char *str The string to be encoded.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *URLEncodePassword(const char *str)
+{
+ int i,j;
+ char *copy=(char*)malloc(3*strlen(str)+1);
+
+ for(i=0,j=0;str[i];i++)
+    if(isalpha(str[i]) ||
+       isdigit(str[i]) ||
+       str[i]=='?' || str[i]=='=' ||
+       str[i]=='$' || str[i]=='-' || str[i]=='_' || str[i]=='.' || str[i]=='+' || str[i]=='!' ||
+       str[i]=='*' || str[i]=='(' || str[i]==')' || str[i]==',' || str[i]=='\'')
+       copy[j++]=str[i];
+    else
+      {
+       unsigned int val=str[i];
+       copy[j]='%';
+       if(val%16>9)
+          copy[j+2]=val%16+'A'-10;
+       else
+          copy[j+2]=val%16+'0';
+       val/=16;
+       if(val%16>9)
+          copy[j+1]=val%16+'A'-10;
+       else
+          copy[j+1]=val%16+'0';
+       j+=3;
+      }
+
+ copy[j]=0;
+
+ return(copy);
+}
+
 
 /*++++++++++++++++++++++++++++++++++++++
   Generate a hash value for a string.
@@ -487,7 +587,8 @@ char *RFC822Date(long t,int utc)
 {
  static char *week[7]={"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
  static char *month[12]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
- static char value[32];
+ static char value[2][32];
+ static int which=0;
  struct tm *tim;
 
  if(utc)
@@ -499,8 +600,10 @@ char *RFC822Date(long t,int utc)
       {tim=gmtime(&t);utc=1;}
    }
 
- /*             Sun, 06 Nov 1994 08:49:37 GMT    ; RFC 822, updated by RFC 1123 */
- sprintf(value,"%3s, %02d %3s %4d %02d:%02d:%02d %s",
+ which^=1;
+
+ /* Sun, 06 Nov 1994 08:49:37 GMT    ; RFC 822, updated by RFC 1123 */
+ sprintf(value[which],"%3s, %02d %3s %4d %02d:%02d:%02d %s",
          week[tim->tm_wday],
          tim->tm_mday,
          month[tim->tm_mon],
@@ -508,9 +611,16 @@ char *RFC822Date(long t,int utc)
          tim->tm_hour,
          tim->tm_min,
          tim->tm_sec,
-         utc?"GMT":tzname[tim->tm_isdst>0]);
+#if defined(__FreeBSD__) || defined(__NetBSD__)
+         utc?"GMT":tim->tm_zone
+#elif defined(__CYGWIN__)
+         utc?"GMT":_tzname[tim->tm_isdst>0]
+#else
+         utc?"GMT":tzname[tim->tm_isdst>0]
+#endif
+         );
 
- return(value);
+ return(value[which]);
 }
 
 
@@ -715,5 +825,3 @@ char* HTMLString(const char* c)
 
  return(ret);
 }
-
-#endif /* URL_ONLY */
