@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/purge.c 2.9 1998/07/05 13:24:02 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/purge.c 2.14 1999/03/17 19:50:28 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.3.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4c.
   Purge old files from the cache.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97 Andrew M. Bishop
+  This file Copyright 1996,97,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -21,6 +21,7 @@
 #include <utime.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/vfs.h>
 #include <time.h>
 #include <dirent.h>
 #include <unistd.h>
@@ -41,8 +42,8 @@ static int PurgeFiles(char *proto,char *host);
 /*+ The current time. +*/
 static time_t now;
 
-/*+ The number of bytes left of each age. +*/
-static int *bytes_by_age;
+/*+ The number of blocks left of each age. +*/
+static int *blocks_by_age;
 
 /*+ The scaling factor for the ages in the second pass. +*/
 static double age_scale;
@@ -59,36 +60,62 @@ int pass;
 
 void PurgeCache(int fd)
 {
- int total_bytes=0,total_dirs=0;
  int i,p;
- int age_for_size=-1;
+ int blocksize,diskfree;
+ struct statfs sbuf;
+ struct stat buf;
 
- now=time(NULL)+60;
+ now=time(NULL)+600;
 
- age_scale=0;
+ age_scale=-1;
 
- bytes_by_age=(int*)malloc((DefaultPurgeAge+1)*sizeof(int));
+ blocks_by_age=(int*)malloc((DefaultPurgeAge+2)*sizeof(int));
 
- for(i=0;i<=DefaultPurgeAge;i++)
-    bytes_by_age[i]=0;
+ for(i=0;i<=DefaultPurgeAge+1;i++)
+    blocks_by_age[i]=0;
 
- if(PurgeUseMTime)
-    write_string(fd,"(Using modification time.)\n");
+ if(stat(".",&buf))
+   {
+    PrintMessage(Warning,"Cannot determine the disk block size [%!s]");
+    blocksize=1024;
+   }
  else
-    write_string(fd,"(Using last access time.)\n");
+    blocksize=buf.st_size/buf.st_blocks;
 
- if(PurgeUseURL)
-    write_string(fd,"(Using the full URL.)\n");
+ if(statfs(".",&sbuf) || sbuf.f_bsize==-1)
+   {
+    PrintMessage(Warning,"Cannot determine the disk free space [%!s]");
+    diskfree=0;
+   }
  else
-    write_string(fd,"(Using the hostname and protocol only.)\n");
+    diskfree=sbuf.f_bsize*sbuf.f_bavail/blocksize;
 
  for(pass=1;pass<=2;pass++)
    {
+    int total_blocks=0,total_dirs=0;
+
+    write_string(fd,"\n");
+
     if(PurgeCacheSize)
        if(pass==1)
-          write_string(fd,"\nPass 1: Checking dates and sizes of files.\n");
+          write_string(fd,"Pass 1: Checking dates and sizes of files:\n");
        else
-          write_string(fd,"\nPass 2: Purging files down to specified size.\n");
+          write_string(fd,"Pass 2: Purging files down to specified size:\n");
+    else
+       write_string(fd,"Checking dates of files:\n");
+
+    if(pass==1)
+      {
+       if(PurgeUseMTime)
+          write_string(fd,"  (Using modification time.)\n");
+       else
+          write_string(fd,"  (Using last access time.)\n");
+
+       if(PurgeUseURL)
+          write_string(fd,"  (Using the full URL.)\n");
+       else
+          write_string(fd,"  (Using the hostname and protocol only.)\n");
+      }
 
     write_string(fd,"\n");
 
@@ -104,6 +131,9 @@ void PurgeCache(int fd)
        if(stat(proto,&buf))
          {PrintMessage(Inform,"Cannot stat directory '%s' [%!s]; not purged",proto);continue;}
 
+       total_blocks+=buf.st_blocks;
+       blocks_by_age[DefaultPurgeAge+1]+=buf.st_blocks;
+
        if(chdir(proto))
          {PrintMessage(Warning,"Cannot change to directory '%s' [%!s]; not purged.",proto);continue;}
 
@@ -126,9 +156,9 @@ void PurgeCache(int fd)
              PrintMessage(Inform,"Cannot stat file '%s/%s' [%!s]; race condition?",proto,ent->d_name);
           else if(S_ISDIR(buf.st_mode))
             {
-             int bytes=PurgeFiles(proto,ent->d_name);
+             int blocks=PurgeFiles(proto,ent->d_name);
 
-             if(bytes==-1)
+             if(blocks==-1)
                {
 #if DO_DELETE
                 if(rmdir(ent->d_name))
@@ -144,19 +174,19 @@ void PurgeCache(int fd)
                 utbuf.actime=buf.st_atime;
                 utbuf.modtime=buf.st_mtime;
                 utime(ent->d_name,&utbuf);
-               }
 
-             if(bytes>0)
-                total_bytes+=bytes;
-             if(bytes>=0)
+                total_blocks+=buf.st_blocks;
+                blocks_by_age[DefaultPurgeAge+1]+=buf.st_blocks;
+                total_blocks+=blocks;
                 total_dirs++;
+               }
 
              if(PurgeUseURL)
                {
-                if(bytes==-1)
-                   write_formatted(fd,"Purged %6s://%-32s ; (empty) - deleted.\n",proto,ent->d_name);
+                if(blocks==-1)
+                   write_formatted(fd,"Purged %6s://%-32s ; (empty) - deleted\n",proto,ent->d_name);
                 else
-                   write_formatted(fd,"Purged %6s://%-32s ; %5d kB.\n",proto,ent->d_name,(bytes+512)/1024);
+                   write_formatted(fd,"Purged %6s://%-32s ; %5ld kB\n",proto,ent->d_name,(blocks+buf.st_blocks)*blocksize/1024);
                }
              else
                {
@@ -166,11 +196,11 @@ void PurgeCache(int fd)
                    age=(int)(age*age_scale);
 
                 if(age<0)
-                   write_formatted(fd,"Not Purged       %6s://%-32s ; %5d kB.\n",proto,ent->d_name,(bytes+512)/1024);
-                else if(bytes==-1)
-                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; (empty) - deleted.\n",age,proto,ent->d_name);
+                   write_formatted(fd,"Not Purged       %6s://%-32s ; %5ld kB\n",proto,ent->d_name,(blocks+buf.st_blocks)*blocksize/1024);
+                else if(blocks==-1)
+                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; (empty) - deleted\n",age,proto,ent->d_name);
                 else
-                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; %5d kB.\n",age,proto,ent->d_name,(bytes+512)/1024);
+                   write_formatted(fd,"Purged (%2d days) %6s://%-32s ; %5ld kB\n",age,proto,ent->d_name,(blocks+buf.st_blocks)*blocksize/1024);
                }
             }
          }
@@ -179,52 +209,120 @@ void PurgeCache(int fd)
        chdir("..");
       }
 
-    write_formatted(fd,"Total of %d directories ; %d kB.\n",total_dirs,(total_bytes+512)/1024);
+    write_string(fd,"\n");
+    write_formatted(fd,"Total of %d directories ; %d kB\n",total_dirs,total_blocks*blocksize/1024);
 
     if(pass==1)
       {
+       int age_for_size=-1,age_for_free=-1;
+       int age_blocks_used=blocks_by_age[DefaultPurgeAge+1];
+       int age_blocks_free=diskfree+total_blocks-blocks_by_age[DefaultPurgeAge+1];
+
        write_string(fd,"\n");
-       write_string(fd,"(All ages scaled to the range 0 -> default age.)\n");
+       write_string(fd,"Age Profile of cached pages:\n");
+       write_string(fd,"  (All ages scaled to the range 0 -> default age.)\n");
        write_string(fd,"\n");
 
-       for(total_bytes=i=0;i<=DefaultPurgeAge;i++)
+       write_formatted(fd,"Total not purged   ; %5d kB (%6d kB free)\n",
+                       age_blocks_used*blocksize/1024,age_blocks_free*blocksize/1024);
+       write_string(fd,"\n");
+
+       for(i=0;i<=DefaultPurgeAge;i++)
          {
-          total_bytes+=bytes_by_age[i];
+          age_blocks_used+=blocks_by_age[i];
+          age_blocks_free-=blocks_by_age[i];
 
-          if(PurgeCacheSize && age_for_size<0 && ((total_bytes+512)/1024)>(1024*PurgeCacheSize))
+          if(PurgeCacheSize && age_for_size<0 && (age_blocks_used*blocksize/1024)>(1024*PurgeCacheSize))
             {
-             if(DefaultPurgeAge)
-               {
-                age_for_size=i;
-                age_scale=(double)age_for_size/(double)DefaultPurgeAge;
-               }
-             else
-                age_scale=0;
-             write_formatted(fd,"Cutoff Age is %2d days for %3d MB.\n",age_for_size,PurgeCacheSize);
+             age_for_size=i;
+
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB cache size\n",age_for_size,PurgeCacheSize);
+            }
+
+          if(PurgeDiskFree && age_for_free<0 && (age_blocks_free*blocksize/1024)<(1024*PurgeDiskFree))
+            {
+             age_for_free=i;
+
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB disk free\n",age_for_free,PurgeDiskFree);
             }
 
           if(i==DefaultPurgeAge)
-             write_formatted(fd,"Total all ages     ; %5d kB.\n",(total_bytes+512)/1024);
+             write_formatted(fd,"Total all ages     ; %5d kB (%6d kB free)\n",
+                             (age_blocks_used-blocks_by_age[DefaultPurgeAge+1])*blocksize/1024,age_blocks_free*blocksize/1024);
           else
-             write_formatted(fd,"Newer than %2d day%c ; %5d kB.\n",i+1,i?'s':' ',(total_bytes+512)/1024);
+             write_formatted(fd,"Newer than %2d day%c ; %5d kB (%6d kB free)\n",i+1,i?'s':' ',
+                             (age_blocks_used-blocks_by_age[DefaultPurgeAge+1])*blocksize/1024,age_blocks_free*blocksize/1024);
          }
-       total_bytes=0;
+
+       if(DefaultPurgeAge)
+         {
+          if(age_for_size!=-1 && (age_for_size<=age_for_free || age_for_free==-1))
+             age_scale=(double)age_for_size/(double)DefaultPurgeAge;
+          else if(age_for_free!=-1 && (age_for_free<age_for_size || age_for_size==-1))
+             age_scale=(double)age_for_free/(double)DefaultPurgeAge;
+         }
+       else if(age_for_size!=-1 || age_for_free!=-1)
+          age_scale=0;
       }
 
-    if(!PurgeCacheSize || age_for_size==-1)
+    if(age_scale==-1)
        break;
    }
 
  write_string(fd,"\n");
 
- free(bytes_by_age);
+ free(blocks_by_age);
+
+ /* Purge the tmp.* files in outgoing. */
+
+ if(chdir("outgoing"))
+    PrintMessage(Warning,"Cannot change to directory 'outgoing' [%!s]; not purged.");
+ else
+   {
+    DIR *dir;
+    struct dirent* ent;
+
+    dir=opendir(".");
+    if(!dir)
+      PrintMessage(Warning,"Cannot open directory 'outgoing' [%!s]; not purged.");
+    else
+      {
+       ent=readdir(dir);  /* skip .  */
+       if(!ent)
+          PrintMessage(Warning,"Cannot read directory 'outgoing' [%!s]; not purged.");
+       else
+         {
+          ent=readdir(dir);  /* skip .. */
+
+          while((ent=readdir(dir)))
+             if(!strncmp(ent->d_name,"tmp.",4))
+               {
+                struct stat buf;
+
+                if(!stat(ent->d_name,&buf) && buf.st_mtime<(now-60))
+                  {
+#if DO_DELETE
+                   if(unlink(ent->d_name))
+                      PrintMessage(Warning,"Cannot unlink file 'outgoing/%s' [%!s].",ent->d_name);
+#else
+                   PrintMessage(Debug,"unlink(outgoing/%s).",ent->d_name);
+#endif
+                  }
+               }
+         }
+
+       closedir(dir);
+      }
+   }
+
+ chdir("..");
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
   Delete the file in the current directory that are older than the specified age.
 
-  int PurgeFiles Returns the number of bytes in files that are left.
+  int PurgeFiles Returns the number of blocks in files that are left.
 
   char *proto The name of the protocol directory to purge.
 
@@ -233,7 +331,7 @@ void PurgeCache(int fd)
 
 static int PurgeFiles(char *proto,char *host)
 {
- int bytes_left=-1;
+ int blocks_left=-1;
  DIR *dir;
  struct dirent* ent;
  int def_age=WhatPurgeAge(proto,host,"/");
@@ -265,6 +363,16 @@ static int PurgeFiles(char *proto,char *host)
        int age=DefaultPurgeAge;
        time_t t=now;
 
+       if(buf.st_mtime>now || buf.st_atime>now)
+         {
+          PrintMessage(Inform,"Cached file '%s/%s/%s' has a future timestamp; changing timestamp.",proto,host,ent->d_name);
+#if DO_DELETE
+          utime(ent->d_name,NULL);
+#else
+          PrintMessage(Debug,"utime(%s/%s/%s).",proto,host,ent->d_name);
+#endif
+         }
+
        if(*ent->d_name=='U' || *ent->d_name=='D')
          {
           int s;
@@ -275,7 +383,7 @@ static int PurgeFiles(char *proto,char *host)
 
           if(s)
             {
-             PrintMessage(Inform,"Cached file '%s' is not complete (U* and D* files); deleting it",ent->d_name);
+             PrintMessage(Inform,"Cached file '%s/%s/%s' is not complete (U* and D* files); deleting it.",proto,host,ent->d_name);
              age=0;
             }
           else if(*ent->d_name=='U')
@@ -303,7 +411,7 @@ static int PurgeFiles(char *proto,char *host)
          }
        else
          {
-          PrintMessage(Inform,"Cached file '%s' is not valid (U* or D* file); deleting it",ent->d_name);
+          PrintMessage(Inform,"Cached file '%s/%s/%s' is not valid (U* or D* file); deleting it.",proto,host,ent->d_name);
           age=0;
          }
 
@@ -312,11 +420,11 @@ static int PurgeFiles(char *proto,char *host)
 
        if(age==-1 || t>(now-age*(24*3600)))
          {
-          int size=buf.st_size+buf2.st_size;
+          int size=buf.st_blocks+buf2.st_blocks;
 
-          if(bytes_left==-1)
-             bytes_left=0;
-          bytes_left+=size;
+          if(blocks_left==-1)
+             blocks_left=0;
+          blocks_left+=size;
 
           if(age>0)
             {
@@ -326,8 +434,10 @@ static int PurgeFiles(char *proto,char *host)
 
              if(days>DefaultPurgeAge)
                 days=DefaultPurgeAge;
-             bytes_by_age[days]+=size;
+             blocks_by_age[days]+=size;
             }
+          else
+             blocks_by_age[DefaultPurgeAge+1]+=size;
          }
        else
          {
@@ -356,5 +466,5 @@ static int PurgeFiles(char *proto,char *host)
  closedir(dir);
  chdir("..");
 
- return(bytes_left);
+ return(blocks_left);
 }
