@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffle-tools.c 1.5 1998/06/08 18:16:07 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffle-tools.c 1.12 1999/02/06 17:31:09 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.2.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4c.
   Tools for use in the cache for version 2.x.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98 Andrew M. Bishop
+  This file Copyright 1997,98,99 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -33,16 +33,24 @@
 #include "errors.h"
 #include "config.h"
 
+/*+ Need this for Win32 to use binary mode +*/
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
 
 #define LS         1
-#define MV         2
-#define RM         3
-#define LS_SPECIAL 4
+#define LS_SPECIAL 2
+#define MV         3
+#define RM         4
+#define READ       5
+#define WRITE      6
 
 static void wwwoffle_ls(URL *Url);
 static void wwwoffle_ls_special(char *name);
 static void wwwoffle_mv(URL *Url1,URL *Url2);
 static void wwwoffle_rm(URL *Url);
+static void wwwoffle_read(URL *Url);
+static void wwwoffle_write(URL *Url);
 
 static void ls(char *file);
 
@@ -71,9 +79,14 @@ int main(int argc,char **argv)
     mode=MV;
  else if(!strcmp(p,"wwwoffle-rm"))
     mode=RM;
+ else if(!strcmp(p,"wwwoffle-read"))
+    mode=READ;
+ else if(!strcmp(p,"wwwoffle-write"))
+    mode=WRITE;
  else
    {
-    fprintf(stderr,"Program must be started as wwwoffle-ls, wwwoffle-mv or wwwoffle-rm\n");
+    fprintf(stderr,"Program must be started as: wwwoffle-ls, wwwoffle-mv, wwwoffle-rm\n");
+    fprintf(stderr,"                            wwwoffle-read or wwwoffle-write\n");
     exit(1);
    }
 
@@ -81,12 +94,16 @@ int main(int argc,char **argv)
 
  if(mode==LS && argc!=2)
    {fprintf(stderr,"Usage: wwwoffle-ls ( <dir>/<subdir> | <protocol>://<host> | <URL> | \n"
-                   "                     outgoing | lasttime )\n");exit(0);}
+                   "                     outgoing | monitor | lasttime | prevtime[0-9] )\n");exit(0);}
  else if(mode==MV && argc!=3)
    {fprintf(stderr,"Usage: wwwoffle-mv (<dir1>/<subdir1> | <protocol1>://<host1>)\n"
                    "                   (<dir2>/<subdir2> | <protocol2>://<host2>)\n");exit(0);}
  else if(mode==RM && argc!=2)
    {fprintf(stderr,"Usage: wwwoffle-rm <URL>\n");exit(0);}
+ else if(mode==READ && argc!=2)
+   {fprintf(stderr,"Usage: wwwoffle-read <URL>\n");exit(0);}
+ else if(mode==WRITE && argc!=2)
+   {fprintf(stderr,"Usage: wwwoffle-write <URL>\n");exit(0);}
 
  if(stat("outgoing",&buf) || !S_ISDIR(buf.st_mode))
    {fprintf(stderr,"The wwwoffle-tools programs must be started from the spool directory\n"
@@ -94,8 +111,6 @@ int main(int argc,char **argv)
 
  InitErrorHandler(p,0,1);
 
- LogLevel=Inform;
-
  /* Get the arguments */
 
  for(i=1;i<argc;i++)
@@ -103,7 +118,8 @@ int main(int argc,char **argv)
     char *colon=strchr(argv[i],':');
     char *slash=strchr(argv[i],'/');
 
-    if(mode==LS && (!strcmp(argv[i],"outgoing") || !strcmp(argv[i],"lasttime")))
+    if(mode==LS && (!strcmp(argv[i],"outgoing") || !strcmp(argv[i],"monitor") ||
+                    !strcmp(argv[i],"lasttime") || (!strncmp(argv[i],"prevtime",8) && isdigit(argv[i][8]))))
       {
        mode=LS_SPECIAL;
       }
@@ -129,12 +145,16 @@ int main(int argc,char **argv)
 
  if(mode==LS)
     wwwoffle_ls(Url[0]);
+ else if(mode==LS_SPECIAL)
+    wwwoffle_ls_special(argv[1]);
  else if(mode==MV)
     wwwoffle_mv(Url[0],Url[1]);
  else if(mode==RM)
     wwwoffle_rm(Url[0]);
- else if(mode==LS_SPECIAL)
-    wwwoffle_ls_special(argv[1]);
+ else if(mode==READ)
+    wwwoffle_read(Url[0]);
+ else if(mode==WRITE)
+    wwwoffle_write(Url[0]);
 
  exit(0);
 }
@@ -178,7 +198,7 @@ static void wwwoffle_ls(URL *Url)
 
     while((ent=readdir(dir)))
       {
-       if(*ent->d_name=='D')
+       if(*ent->d_name=='D' && ent->d_name[strlen(ent->d_name)-1]!='~')
           ls(ent->d_name);
       }
 
@@ -190,7 +210,7 @@ static void wwwoffle_ls(URL *Url)
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  List the URLs within the outgoing or lasttime special directory of the cache.
+  List the URLs within the outgoing, monitor or lasttime/prevtime special directory of the cache.
 
   char *name The name of the directory to list.
   ++++++++++++++++++++++++++++++++++++++*/
@@ -215,7 +235,7 @@ static void wwwoffle_ls_special(char *name)
 
  while((ent=readdir(dir)))
    {
-    if(*ent->d_name=='D' || *ent->d_name=='O')
+    if((*ent->d_name=='D' || *ent->d_name=='O') && ent->d_name[strlen(ent->d_name)-1]!='~')
        ls(ent->d_name);
    }
 
@@ -308,8 +328,7 @@ static void wwwoffle_mv(URL *Url1,URL *Url2)
           sprintf(url2,"%s://%s%s",Url2->proto,Url2->host,Url->pathp);
           FreeURL(Url);
 
-          name1=(char*)malloc(strlen(ent->d_name)+1);
-          strcpy(name1,ent->d_name);
+          name1=ent->d_name;
 
           Url=SplitURL(url2);
           name2=URLToFileName(Url);
@@ -328,14 +347,13 @@ static void wwwoffle_mv(URL *Url1,URL *Url2)
 
           *name1=*name2='U';
           sprintf(path2,"../../%s/%s/%s",Url2->proto,Url2->host,name2);
-          fd2=open(path2,O_WRONLY|O_CREAT|O_TRUNC);
+          fd2=open(path2,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY);
           write_string(fd2,Url->name);
           close(fd2);
           unlink(name2);
 
           free(url1);
           free(url2);
-          free(name1);
           free(name2);
           free(path2);
           FreeURL(Url);
@@ -359,3 +377,49 @@ static void wwwoffle_rm(URL *Url)
 {
  DeleteWebpageSpoolFile(Url,0);
 }
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Read a URL and output on stdout.
+
+  URL *Url The URL to read.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void wwwoffle_read(URL *Url)
+{
+ char buffer[256];
+ int n,spool=OpenWebpageSpoolFile(1,Url);
+
+ if(spool==-1)
+    return;
+
+ init_buffer(spool);
+
+ while((n=read_data(spool,buffer,256))>0)
+    write_data(1,buffer,n);
+
+ close(spool);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  Write a URL from the input on stdin.
+
+  URL *Url The URL to write.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void wwwoffle_write(URL *Url)
+{
+ char buffer[256];
+ int n,spool=OpenWebpageSpoolFile(0,Url);
+
+ if(spool==-1)
+    return;
+
+ init_buffer(0);
+
+ while((n=read_data(0,buffer,256))>0)
+    write_data(spool,buffer,n);
+
+ close(spool);
+}
