@@ -1,7 +1,7 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.19 1998/06/21 09:08:54 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.24 1999/01/26 17:42:01 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.2a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4b.
   A user level program to interact with the server.
   ******************/ /******************
   Written by Andrew M. Bishop
@@ -23,6 +23,7 @@
 
 #include "version.h"
 #include "wwwoffle.h"
+#include "document.h"
 #include "misc.h"
 #include "config.h"
 #include "sockets.h"
@@ -50,6 +51,7 @@ typedef enum _Action
 Action;
 
 
+static void add_url_list(char **links);
 static void add_url_file(char *url_file);
 
 /*+ The list of URLs or files. +*/
@@ -444,11 +446,29 @@ int main(int argc, char** argv)
 
           printf("Getting: %s\n",Url->name);
 
-          write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
-                                 "Pragma: wwwoffle\r\n"
-                                 "Accept: */*\r\n"
-                                 "\r\n",
-                          refresh,Url->name);
+          if(Url->user)
+            {
+             char *userpass1=(char*)(malloc(strlen(Url->user)+(Url->pass?strlen(Url->pass):1)+3)),*userpass2;
+
+             sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
+             userpass2=Base64Encode(userpass1,strlen(userpass1));
+
+             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
+                                    "Pragma: wwwoffle\r\n"
+                                    "Accept: */*\r\n"
+                                    "Authorization: Basic %s\r\n"
+                                    "\r\n",
+                             refresh,Url->name,userpass2);
+
+             free(userpass1);
+             free(userpass2);
+            }
+          else
+             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
+                                    "Pragma: wwwoffle\r\n"
+                                    "Accept: */*\r\n"
+                                    "\r\n",
+                             refresh,Url->name);
 
           while(read_data(socket,buffer256,256)>0)
              ;
@@ -462,14 +482,12 @@ int main(int argc, char** argv)
           int file;
           char *buffer=(char*)malloc(strlen(url_file_list[i])+256);
           char **links;
-          int j;
 
           if(strcmp(url_file_list[i],"-"))
             {
              file=open(url_file_list[i],O_RDONLY);
              if(file==-1)
                {PrintMessage(Warning,"Cannot open file '%s' for reading.",url_file_list[i]);continue;}
-             init_buffer(file);
 
              printf("Reading: %s\n",url_file_list[i]);
             }
@@ -480,6 +498,8 @@ int main(int argc, char** argv)
              printf("Reading: <stdin>\n");
             }
 
+          init_buffer(file);
+
           strcpy(buffer,"file://localhost");
           if(*url_file_list[i]=='/')
              strcat(buffer,url_file_list[i]);
@@ -496,40 +516,28 @@ int main(int argc, char** argv)
             }
 
           Url=SplitURL(buffer);
-          ParseHTML(file,Url,1);
+          ParseHTML(file,Url);
 
-          if(images && (links=ListImages()))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
+          if(stylesheets && (links=GetReferences(RefStyleSheet)))
+             add_url_list(links);
 
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
+          if(images && (links=GetReferences(RefImage)))
+             add_url_list(links);
 
-                FreeURL(linkUrl);
-               }
+          if(frames && (links=GetReferences(RefFrame)))
+             add_url_list(links);
 
-          if((links=ListLinks()))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
+          if(scripts && (links=GetReferences(RefScript)))
+             add_url_list(links);
 
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
+          if(objects && (links=GetReferences(RefObject)))
+             add_url_list(links);
 
-                FreeURL(linkUrl);
-               }
+          if(objects && (links=GetReferences(RefInlineObject)))
+             add_url_list(links);
 
-          if(frames && (links=ListFrames()))
-             for(j=0;links[j];j++)
-               {
-                URL *linkUrl=SplitURL(links[j]);
-
-                if(strcmp(linkUrl->proto,"file"))
-                   add_url_file(linkUrl->name);
-
-                FreeURL(linkUrl);
-               }
+          if((links=GetReferences(RefLink)))
+             add_url_list(links);
 
           FreeURL(Url);
           if(file!=0)
@@ -546,7 +554,7 @@ int main(int argc, char** argv)
     int nbytes;
 
     if(!n_url_file_list)
-       PrintMessage(Fatal,"No URL specified to output");
+       PrintMessage(Fatal,"No URL specified to output.");
 
     socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port);
     init_buffer(socket);
@@ -559,11 +567,29 @@ int main(int argc, char** argv)
     if(action!=Output && action!=OutputWithHeader)
        fprintf(stderr,"Getting: %s\n",Url->name);
 
-    write_formatted(socket,"GET %s HTTP/1.0\r\n"
-                           "Pragma: wwwoffle\r\n"
-                           "Accept: */*\r\n"
-                           "\r\n",
-                    Url->name);
+    if(Url->user)
+      {
+       char *userpass1=(char*)(malloc(strlen(Url->user)+(Url->pass?strlen(Url->pass):1)+3)),*userpass2;
+
+       sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
+       userpass2=Base64Encode(userpass1,strlen(userpass1));
+
+       write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                              "Pragma: wwwoffle\r\n"
+                              "Accept: */*\r\n"
+                              "Authorization: Basic %s\r\n"
+                              "\r\n",
+                       Url->name,userpass2);
+
+       free(userpass1);
+       free(userpass2);
+      }
+    else
+       write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                              "Pragma: wwwoffle\r\n"
+                              "Accept: */*\r\n"
+                              "\r\n",
+                       Url->name);
 
     if((buffer=read_line(socket,buffer)))
       {
@@ -624,8 +650,8 @@ static void usage(int verbose)
 
  if(verbose)
     fprintf(stderr,
-            "(c) Andrew M. Bishop 1996,97,98 [       amb@gedanken.demon.co.uk ]\n"
-            "                                [http://www.gedanken.demon.co.uk/]\n"
+            "(c) Andrew M. Bishop 1996,97,98,99 [       amb@gedanken.demon.co.uk ]\n"
+            "                                   [http://www.gedanken.demon.co.uk/]\n"
             "\n");
 
  fprintf(stderr,
@@ -699,6 +725,28 @@ static void usage(int verbose)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Add a list of URLs to the list.
+
+  char **links The list of URLs to add.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void add_url_list(char **links)
+{
+ int i;
+
+ for(i=0;links[i];i++)
+   {
+    URL *linkUrl=SplitURL(links[i]);
+
+    if(strcmp(linkUrl->proto,"file"))
+       add_url_file(linkUrl->name);
+
+    FreeURL(linkUrl);
+   }
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Add a URL or a file to the list.
 
