@@ -1,7 +1,7 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 2.84 1998/09/03 17:32:05 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 2.113 1999/02/20 12:13:07 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.3b.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4c.
   A server to fetch the required pages.
   ******************/ /******************
   Written by Andrew M. Bishop
@@ -25,6 +25,7 @@
 #include <fcntl.h>
 
 #include "wwwoffle.h"
+#include "document.h"
 #include "misc.h"
 #include "proto.h"
 #include "config.h"
@@ -32,6 +33,7 @@
 #include "errors.h"
 
 
+static void ssl_tunnel(int client,URL *Url,char *request_head);
 static int request_new(URL *Url,char *referer);
 static void uninstall_sighandlers(void);
 
@@ -79,6 +81,8 @@ int wwwoffles(int online,int browser,int client)
  URL *Url,*Urlpw=NULL;
  Mode mode=None;
  int lasttime_exists=0,spool_exists=0,spool_exists_pw=0;
+ int offline_request=1;
+ char *proxy_auth,*proxy_user;
 
  /* Initialise things. */
 
@@ -126,6 +130,8 @@ int wwwoffles(int online,int browser,int client)
  else /* mode==Fetch */
     url=ParseRequest(outgoing,&request_head,&request_body);
 
+ PrintMessage(ExtraDebug,"Incoming Request Head (from browser)\n%s",request_head?request_head:"(empty)");
+
  if(!url)
    {
     PrintMessage(Warning,"Could not parse HTTP request (%s).",request_head?"Parse error":"Empty request");
@@ -138,10 +144,72 @@ int wwwoffles(int online,int browser,int client)
 
  Url=SplitURL(url);
 
- PrintMessage(Inform,"URL='%s'%s.",Url->name,Url->user?" (With username/password)":"");
+ /* Authenticate the user with the proxy */
+
+ if(!(proxy_user=IsAllowedConnectUser(proxy_auth=GetHTTPHeader(request_head,"Proxy-Authorization:"))))
+   {
+    PrintMessage(Inform,"HTTP Proxy connection rejected from unauthenticated user."); /* Used in audit-usage.pl */
+    HTMLMessageHead(client,407,"WWWOFFLE Proxy Authentication Required",
+                    "Proxy-Authenticate","Basic realm=\"wwwoffle-proxy\"",
+                    NULL);
+    HTMLMessageBody(client,"ProxyAuthFail",
+                    NULL);
+    exit(1);
+   }
+ else if(proxy_auth)
+    PrintMessage(Inform,"HTTP Proxy connection from user '%s'.",proxy_user); /* Used in audit-usage.pl */
+
+ /* Check the HTTP method that is requested */
+
+ if(!strncasecmp(request_head,"CONNECT",3))
+   {
+    PrintMessage(Inform,"SSL='%s'.",Url->host); /* Used in audit-usage.pl */
+
+    if(mode==Real || mode==SpoolOrReal)
+      {
+       if(IsSSLAllowedPort(Url->host))
+         {
+          ssl_tunnel(client,Url,request_head);
+          exit(0);
+         }
+       else
+         {
+          PrintMessage(Warning,"A SSL proxy connection for %s was received but is not allowed.",Url->host);
+          if(client!=-1)
+             HTMLMessage(client,500,"WWWOFFLE Server Error",NULL,"ServerError",
+                         "error","SSL proxy connection to specified port is not allowed.",
+                         NULL);
+         }
+      }
+    else
+      {
+       PrintMessage(Warning,"A SSL proxy connection for %s was received but wwwoffles is in wrong mode.",Url->host);
+       if(client!=-1)
+          HTMLMessage(client,500,"WWWOFFLE Server Error",NULL,"ServerError",
+                      "error","SSL proxy connection while offline is not allowed.",
+                      NULL);
+      }
+    exit(1);
+   }
+ else if(strncasecmp(request_head,"POST",4) && strncasecmp(request_head,"GET",3) && strncasecmp(request_head,"HEAD",4))
+   {
+    char *method=request_head,*m=method;
+
+    while(*m!=' ')
+       m++;
+    *m=0;
+
+    PrintMessage(Warning,"The requested HTTP method '%s' is not supported.",method);
+    if(client!=-1)
+       HTMLMessage(client,501,"WWWOFFLE Method Unsupported",NULL,"MethodUnsupported",
+                   "method",method,
+                   NULL);
+    exit(1);
+   }
+
+ PrintMessage(Inform,"URL='%s'%s.",Url->name,Url->user?" (With username/password)":""); /* Used in audit-usage.pl */
  PrintMessage(Debug,"proto='%s'; host='%s'; path='%s'; args='%s'; user:pass='%s:%s'.",
               Url->proto,Url->host,Url->path,Url->args,Url->user,Url->pass);
- PrintMessage(ExtraDebug,"Incoming Request Head (from browser)\n%s",request_head);
 
  /* Check for an alias. */
 
@@ -241,37 +309,25 @@ int wwwoffles(int online,int browser,int client)
              exit(0);
             }
 
-          if(recurse && Url->args && *Url->args=='!')
+          if(recurse)
             {
-             if(mode==Spool || mode==Real || mode==SpoolOrReal)
-               {
-                mode=SpoolGet;
-                free(request_head);
-                request_head=RequestURL(newUrl,NULL);
-                if(request_body)
-                   free(request_body);
-                request_body=NULL;
-               }
              fetch_again=1;
-            }
-          else if(recurse)
-            {
-             if(mode==Fetch)
-                fetch_again=1;
-             else if(mode==Spool)
-               {
-                mode=SpoolGet;
-                free(newurl);
-                FreeURL(newUrl);
-                newurl=(char*)malloc(strlen(url)+1);
-                strcpy(newurl,url);
-                newUrl=SplitURL(newurl);
-                if(Url->user)
-                   AddURLPassword(newUrl,Url->user,Url->pass);
-               }
-             else if(mode==Real || mode==SpoolOrReal)
+
+             if(mode!=Fetch)
                {
                 mode=SpoolGet;
+
+                if(!Url->args || *Url->args!='!')
+                  {
+                   free(newurl);
+                   FreeURL(newUrl);
+                   newurl=(char*)malloc(strlen(url)+1);
+                   strcpy(newurl,url);
+                   newUrl=SplitURL(newurl);
+                   if(Url->user)
+                      AddURLPassword(newUrl,Url->user,Url->pass);
+                  }
+
                 free(request_head);
                 request_head=RequestURL(newUrl,NULL);
                 if(request_body)
@@ -304,11 +360,6 @@ int wwwoffles(int online,int browser,int client)
        PrintMessage(Inform,"The request to fetch a page from the local host is ignored.");
        exit(0);
       }
-    else if(!strchr(Url->path+1,'/') && !Url->args)
-      {
-       LocalPage(client,Url->path,request_head);
-       exit(0);
-      }
     else if(!strncmp("/index/",Url->path,7))
       {
        IndexPage(client,Url);
@@ -324,11 +375,16 @@ int wwwoffles(int online,int browser,int client)
        MonitorPage(client,Url,request_body);
        exit(0);
       }
-    else if(!strncmp("/local/",Url->path,7) && !Url->args)
+    else if((!strchr(Url->path+1,'/') || !strncmp("/local/",Url->path,7)) && !Url->args)
       {
        LocalPage(client,Url->path,request_head);
        exit(0);
       }
+    else if(!strncmp("/htdig/",Url->path,7))
+      {
+       HTDigPage(client,Url,request_head);
+       exit(0);
+      }
     else
       {
        int i;
@@ -346,6 +402,7 @@ int wwwoffles(int online,int browser,int client)
 
        if(i==NProtocols)
          {
+          PrintMessage(Inform,"The requested URL '%s' does not exist on the local server.",Url->pathp);
           HTMLMessage(client,404,"WWWOFFLE Page Not Found",NULL,"PageNotFound",
                       "url",Url->name,
                       NULL);
@@ -462,6 +519,8 @@ realpassword:
     else
       {
        PrintMessage(Inform,"The request to fetch a page from the local network host '%s' is ignored.",Url->host);
+       if(client!=-1)
+          write_formatted(client,"Cannot fetch %s [On local network]\n",Url->name);
        exit(0);
       }
    }
@@ -490,6 +549,8 @@ realpassword:
     else if(mode==Fetch)
       {
        PrintMessage(Inform,"It is not possible to fetch a URL that was posted.");
+       if(client!=-1)
+          write_formatted(client,"Cannot fetch %s [Reply from a POST]\n",Url->name);
        exit(0);
       }
    }
@@ -538,12 +599,12 @@ realpassword:
    }
  else if(mode==Fetch && ExistsLockWebpageSpoolFile(Url))
    {
-    PrintMessage(Debug,"Already fetching URL");
+    PrintMessage(Debug,"Already fetching URL.");
     if(client!=-1)
-       write_formatted(client,"Fetching %s [Already fetching]\n",Url->name);
+       write_formatted(client,"Cannot fetch %s [Already fetching]\n",Url->name);
     exit(0);
    }
- else if(PragmaNoCache && GetHTTPHeader(request_head,"Pragma: no-cache"))
+ else if(RefreshForced() || (PragmaNoCache && GetHTTPHeader(request_head,"Pragma: no-cache")))
    {
     if(mode==Spool || mode==SpoolGet)
       {
@@ -572,8 +633,18 @@ realpassword:
     spool=OpenWebpageSpoolFile(1,Url);
     init_buffer(spool);
 
-    if(RequestChanged>=0 && RequestChanges(spool,&request_head)==1)
+    if((RequestChanged>=0 || RequestChangedOnce) && RequestChanges(spool,&request_head)==1)
        TouchWebpageSpoolFile(Url);
+    else if(fetch_again)
+      {
+       lseek(spool,0,SEEK_SET);
+       init_buffer(spool);
+
+       if(ParseDocument(spool,Url))
+          RecurseFetch(Url,0);
+
+       exit(4);
+      }
     else
        exit(0);
 
@@ -584,7 +655,7 @@ realpassword:
     spool=OpenWebpageSpoolFile(1,Url);
     init_buffer(spool);
 
-    if(RequestChanged>=0 && RequestChanges(spool,&request_head)==1)
+    if((RequestChanged>=0 || RequestChangedOnce) && RequestChanges(spool,&request_head)==1)
        TouchWebpageSpoolFile(Url);
     else
        mode=Spool;
@@ -599,13 +670,18 @@ realpassword:
        mode=Real;
    }
 
+ /* Don't let htdig request any URLs. */
+
+ if((mode==SpoolGet || mode==Real) && GetHTTPHeader(request_head,"User-Agent: htdig/"))
+   {
+    PrintMessage(Inform,"URL unavailable to be searched.");
+    HTMLMessageHead(client,404,"WWWOFFLE Not Searched",
+                    NULL);
+    exit(0);
+   }
+
  /* mode = Spool, SpoolGet, SpoolWillGet, SpoolPragma, SpoolRefresh, Real, RealRefresh, RealNoCache or Fetch */
 
- /* Create the last time file. */
-
- if(mode==Real || mode==Fetch)
-    lasttime_exists=CreateLastTimeSpoolFile(Url);
-
  /* Set up the file descriptor for the spool file. */
 
  if(mode==Real || mode==Fetch)
@@ -614,9 +690,11 @@ realpassword:
        CreateBackupWebpageSpoolFile(Url);
 
     spool=OpenWebpageSpoolFile(0,Url);
+    CreateLockWebpageSpoolFile(Url);
 
     if(spool==-1)
       {
+       DeleteLockWebpageSpoolFile(Url);
        PrintMessage(Warning,"Cannot open the spooled web page to write.");
        if(mode==Real)
           HTMLMessage(client,500,"WWWOFFLE Server Error",NULL,"ServerError",
@@ -625,7 +703,7 @@ realpassword:
        exit(1);
       }
 
-    CreateLockWebpageSpoolFile(Url);
+    lasttime_exists=CreateLastTimeSpoolFile(Url);
    }
  else if(mode==Spool || mode==SpoolPragma)
    {
@@ -644,7 +722,9 @@ realpassword:
 
  /* Set up the outgoing file. */
 
- if((mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma) && OfflineRequests)
+ if(((offline_request=!IsNotRequestedOffline(Url->proto,Url->host,Url->path)) || online) &&
+    ((mode==SpoolGet && (!ConfirmRequests || Url->local || (Url->args && *Url->args=='!'))) ||
+     mode==SpoolRefresh || mode==SpoolPragma))
    {
     outgoing=OpenOutgoingSpoolFile(0);
 
@@ -681,7 +761,7 @@ realpassword:
                       "url",Url->name,
                       "reason",err,
                       "cache","yes",
-                      "backup",spool_exists?"yes":"",
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
           DeleteLockWebpageSpoolFile(Url);
          }
@@ -690,7 +770,7 @@ realpassword:
                       "url",Url->name,
                       "reason",err,
                       "cache","",
-                      "backup",spool_exists?"yes":"",
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
        exit(1);
       }
@@ -703,6 +783,9 @@ realpassword:
 
  /* Write request to remote server or outgoing file. */
 
+ if(mode==Fetch && client!=-1)
+    write_formatted(client,"Fetching %s ...\n",Url->name);
+
  if(mode==Real || mode==RealNoCache || mode==Fetch)
    {
     char *err=(Url->Protocol->request)(Url,request_head,request_body);
@@ -717,7 +800,7 @@ realpassword:
                       "url",Url->name,
                       "reason",err,
                       "cache","yes",
-                      "backup",spool_exists?"yes":"",
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
           DeleteLockWebpageSpoolFile(Url);
          }
@@ -726,21 +809,23 @@ realpassword:
                       "url",Url->name,
                       "reason",err,
                       "cache","",
-                      "backup",spool_exists?"yes":"",
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
        else if(mode==RealNoCache)
           HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
                       "url",Url->name,
                       "reason",err,
-                      "cache","",
-                      "backup","",
+                      "cache",NULL,
+                      "backup",NULL,
                       NULL);
        else if(mode==Fetch && client!=-1)
-          write_formatted(client,"Fetching %s [Server Connection Error]\n",Url->name);
+          write_formatted(client,"Fetch Failure %s [Server Connection Error]\n",Url->name);
        exit(1);
       }
    }
- else if((mode==SpoolGet || mode==SpoolRefresh || mode==SpoolPragma) && OfflineRequests)
+ else if((offline_request || online) &&
+         ((mode==SpoolGet && (!ConfirmRequests || Url->local || (Url->args && *Url->args=='!'))) ||
+          mode==SpoolRefresh || mode==SpoolPragma))
    {
     int err=write_string(outgoing,request_head);
 
@@ -761,7 +846,7 @@ realpassword:
 
  if(mode==Real || mode==RealNoCache || mode==Fetch)
    {
-    reply_status=ParseReply(Url,&reply_head);
+    reply_status=ParseReply(-1,Url,&reply_head);
 
     if(!reply_head)
       {
@@ -774,7 +859,7 @@ realpassword:
                       "url",Url->name,
                       "reason","TimeoutReply",
                       "cache","yes",
-                      "backup",spool_exists?"yes":"",
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
           DeleteLockWebpageSpoolFile(Url);
          }
@@ -782,11 +867,11 @@ realpassword:
           HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
                       "url",Url->name,
                       "reason","TimeoutReply",
-                      "cache","",
-                      "backup",spool_exists?"yes":"",
+                      "cache",NULL,
+                      "backup",spool_exists?"yes":NULL,
                       NULL);
        else if(mode==Fetch && client!=-1)
-          write_formatted(client,"Fetching %s [Server Reply Error]\n",Url->name);
+          write_formatted(client,"Fetch Failure %s [Server Reply Error]\n",Url->name);
        exit(1);
       }
 
@@ -797,7 +882,7 @@ realpassword:
        char *new_url=MovedLocation(Url,reply_head);
 
        if(client!=-1)
-          write_formatted(client,"Fetching %s [Page Moved]\n",Url->name);
+          write_formatted(client,"Fetching More %s [Page Moved]\n",Url->name);
 
        if(!new_url)
           PrintMessage(Warning,"Cannot parse the reply for the new location.");
@@ -819,8 +904,22 @@ realpassword:
        if(mode==Fetch)
          {
           if(client!=-1)
-             write_formatted(client,"Fetching %s [Page Unchanged]\n",Url->name);
-          exit(0);
+             write_formatted(client,"Not fetching %s [Page Unchanged]\n",Url->name);
+
+          if(fetch_again)
+            {
+             spool=OpenWebpageSpoolFile(1,Url);
+             init_buffer(spool);
+
+             if(ParseDocument(spool,Url))
+                RecurseFetch(Url,1);
+
+             close(spool);
+
+             exit(4);
+            }
+          else
+             exit(0);
          }
        else if(mode==Real)
          {
@@ -834,17 +933,20 @@ realpassword:
       {
        if(request_new(Urlpw,Url->name))
           fetch_again++;
+       if(client!=-1)
+          write_formatted(client,"Fetching More %s [URL with password]\n",Url->name);
       }
     else if(mode==Real && reply_status==401 && Urlpw)
       {
        mode=RealPassword;
       }
-    else if(mode==Fetch && client!=-1)
-       write_formatted(client,"Fetching %s\n",Url->name);
    }
  else
     reply_head=NULL;
 
+ if(mode==Spool || mode==SpoolPragma)
+    reply_status=ParseReply(spool,Url,&reply_head);
+
  reply_body=(char*)malloc(257);
 
  /* mode = Spool, SpoolGet, SpoolWillGet, SpoolPragma, SpoolRefresh, Real, RealRefresh, RealNoCache, RealPassword or Fetch */
@@ -888,23 +990,25 @@ realpassword:
 
        if(err==-1)
          {
-          PrintMessage(Warning,"Error writing to client [%!s]; client disconnected?.");
-          HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
-                      "url",Url->name,
-                      "reason","ClientClose",
-                      "cache","yes",
-                      "backup",spool_exists?"yes":"",
-                      NULL);
+          PrintMessage(Warning,"Error writing to client [%!s]; client disconnected?");
+          if(online!=-1) /* !autodial */
+             HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                         "url",Url->name,
+                         "reason","ClientClose",
+                         "cache","yes",
+                         "backup",spool_exists?"yes":NULL,
+                         NULL);
          }
        else if(n<0)
          {
           PrintMessage(Warning,"Timed out while reading from remote host.");
-          HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
-                      "url",Url->name,
-                      "reason","TimeoutTransfer",
-                      "cache","yes",
-                      "backup",spool_exists?"yes":"",
-                      NULL);
+          if(online!=-1) /* !autodial */
+             HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                         "url",Url->name,
+                         "reason","TimeoutTransfer",
+                         "cache","yes",
+                         "backup",spool_exists?"yes":NULL,
+                         NULL);
          }
        else
           if(spool_exists)
@@ -934,12 +1038,19 @@ realpassword:
                    "url",Url->name,
                    "reason","TimeoutTransfer",
                    "cache","yes",
-                   "backup",spool_exists?"yes":"",
+                   "backup",spool_exists?"yes":NULL,
                    NULL);
+
+       if(client!=-1)
+          write_formatted(client,"Fetch Failure %s [Timeout]\n",Url->name);
       }
     else
+      {
        if(spool_exists)
           DeleteBackupWebpageSpoolFile(Url);
+       if(client!=-1)
+          write_formatted(client,"Fetch Success %s\n",Url->name);
+      }
 
     DeleteLockWebpageSpoolFile(Url);
 
@@ -951,26 +1062,32 @@ realpassword:
        lseek(spool,0,SEEK_SET);
        init_buffer(spool);
 
-       if(ParseHTML(spool,Url,0))
-          fetch_again+=RecurseFetch(Url);
+       if(ParseDocument(spool,Url))
+         {
+          int links=RecurseFetch(Url,1);
+
+          if(client!=-1 && links)
+             write_formatted(client,"Fetching More %s [%d Extra URLs]\n",Url->name,links);
+
+          fetch_again+=links;
+         }
       }
    }
  else if(mode==Spool || mode==SpoolPragma)
    {
     struct stat buf;
     char *head="HTTP/1.0 503 WWWOFFLE Remote Host Error\r\n";
-    int n;
 
     if(ExistsLockWebpageSpoolFile(Url))
       {
-       int t=SocketTimeout;
+       int t=SocketTimeout/6;
 
        PrintMessage(Inform,"Waiting for the page to be unlocked.");
 
-       while((t-=5) && ExistsLockWebpageSpoolFile(Url))
-          sleep(5);
+       while(--t>0 && ExistsLockWebpageSpoolFile(Url))
+          sleep(1);
 
-       if(t==0)
+       if(t<=0)
          {
           PrintMessage(Inform,"Timed out waiting for the page to be unlocked.");
           HTMLMessage(spool,503,"WWWOFFLE File Locked",NULL,"FileLocked",
@@ -980,98 +1097,48 @@ realpassword:
          }
       }
 
-    n=read_data(spool,reply_body,64);
-
-    if(!strncmp(reply_body,head,strlen(head)) || (!fstat(spool,&buf) && buf.st_size==0))
+    if((reply_head && !strncmp(reply_head,head,strlen(head))) || (!fstat(spool,&buf) && buf.st_size==0))
       {
        DeleteWebpageSpoolFile(Url,0);
        RestoreBackupWebpageSpoolFile(Url);
       }
 
-    if(AddInfoRefresh && !GetHTTPHeader(request_head,"Pragma: wwwoffle"))
+    if(EnableHTMLModifications &&
+       reply_status>=200 && reply_status<400 && GetHTTPHeader(reply_head,"Content-Type: text/html") &&
+       !GetHTTPHeader(request_head,"Pragma: wwwoffle") && !GetHTTPHeader(request_head,"User-Agent: htdig/"))
       {
-       struct stat buf;
-
        lseek(spool,0,SEEK_SET);
        init_buffer(spool);
 
-       if(ParseHTML(spool,Url,0) && !fstat(spool,&buf))
-         {
-          time_t t_ago=time(NULL)-buf.st_mtime;
-          int position=GetHTMLEnd();
-          char *date=RFC822Date(buf.st_mtime,0),*timeunit,timeago[8];
-
-          if(t_ago<0)
-            {strcpy(timeago,"?");timeunit="";}
-          else if(t_ago<3600)
-            {sprintf(timeago,"%ld",t_ago/60);timeunit=(t_ago/60==1?"min":"mins");}
-          else if(t_ago<(24*3600))
-            {sprintf(timeago,"%ld",t_ago/3600);timeunit=(t_ago/3600==1?"hour":"hours");}
-          else if(t_ago<(14*24*3600))
-            {sprintf(timeago,"%ld",t_ago/(24*3600));timeunit=(t_ago/(24*3600)==1?"day":"days");}
-          else if(t_ago<(30*24*3600))
-            {sprintf(timeago,"%ld",t_ago/(7*24*3600));timeunit=(t_ago/(7*24*3600)==1?"week":"weeks");}
-          else
-            {sprintf(timeago,"%ld",t_ago/(30*24*3600));timeunit=(t_ago/(30*24*3600)==1?"month":"months");}
-
-          lseek(spool,0,SEEK_SET);
-          init_buffer(spool);
-
-          while((reply_head=read_line(spool,reply_head)))
-            {
-             if(strncasecmp(reply_head,"Content-Length:",15))
-                write_string(client,reply_head);
-             if(reply_head[0]=='\r' || reply_head[0]=='\n')
-                break;
-            }
-
-          while(position>0 && (n=read_data(spool,reply_body,position>256?256:position))>0)
-            {
-             write_data(client,reply_body,n);
-             position-=n;
-            }
-
-          HTMLMessageBody(client,"AddInfoRefresh",
-                          "url",Url->name,
-                          "date",date,
-                          "time",timeago,
-                          "unit",timeunit,
-                          NULL);
-
-          while((n=read_data(spool,reply_body,256))>0)
-             write_data(client,reply_body,n);
-         }
-       else
-         {
-          lseek(spool,0,SEEK_SET);
-          init_buffer(spool);
-
-          while((n=read_data(spool,reply_body,256))>0)
-             write_data(client,reply_body,n);
-         }
+       OutputHTMLWithModifications(client,spool,Url);
       }
     else
-       do
-         {
+      {
+       int n;
+
+       if(reply_head)
+          write_string(client,reply_head);
+
+       while((n=read_data(spool,reply_body,256))>0)
           write_data(client,reply_body,n);
-         }
-       while((n=read_data(spool,reply_body,256))>0);
+      }
    }
  else if(mode==SpoolGet)
    {
-    if(OfflineRequests)
+    if((offline_request || online))
       {
-       if(ConfirmRequests)
-          HTMLMessage(client,200,"WWWOFFLE Confirm Request",NULL,"ConfirmRequest",
+       if(ConfirmRequests && !Url->local && (!Url->args || *Url->args!='!'))
+          HTMLMessage(client,404,"WWWOFFLE Confirm Request",NULL,"ConfirmRequest",
                       "url",Url->name,
                       NULL);
        else if(fetch_again)
           HTMLMessage(client,404,"WWWOFFLE Refresh Will Get",NULL,"RefreshWillGet",
                       "url",Url->name,
+                      "password",HashOutgoingSpoolFile(Url),
                       NULL);
        else
           HTMLMessage(client,404,"WWWOFFLE Will Get",NULL,"WillGet",
-                      "already","",
+                      "already",NULL,
                       "url",Url->name,
                       "password",HashOutgoingSpoolFile(Url),
                       NULL);
@@ -1108,10 +1175,16 @@ realpassword:
  if(spool>=0)
     close(spool);
 
+ if(reply_head)
+    free(reply_head);
+ if(reply_body)
+    free(reply_body);
+
  if(mode==RealPassword)
    {
     FreeURL(Url);Url=Urlpw;Urlpw=NULL;
     spool_exists=spool_exists_pw;spool_exists_pw=0;
+    reply_head=reply_body=NULL;
     mode=Real;
     goto realpassword;
    }
@@ -1121,11 +1194,6 @@ realpassword:
  if(request_body)
     free(request_body);
 
- if(reply_head)
-    free(reply_head);
- if(reply_body)
-    free(reply_body);
-
  if(client>=0)
     CloseSocket(client);
 
@@ -1136,6 +1204,63 @@ realpassword:
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Make an SSL proxy connection.
+
+  int client The client socket.
+
+  URL *Url The URL to get (used for host only).
+
+  char *request_head The head of the request.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void ssl_tunnel(int client,URL *Url,char *request_head)
+{
+ char *err=SSL_Open(Url);
+
+ if(err && ConnectRetry)
+   {
+    PrintMessage(Inform,"Waiting to try connection again.");
+    sleep(10);
+    err=SSL_Open(Url);
+   }
+
+ if(err)
+   {
+    HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                "url",Url->host,
+                "reason",err,
+                "cache",NULL,
+                "backup",NULL,
+                NULL);
+    exit(1);
+   }
+
+ request_head=ModifyRequest(Url,request_head);
+
+ err=SSL_Request(client,Url,request_head);
+
+ if(err)
+   {
+    HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                "url",Url->host,
+                "reason",err,
+                "cache",NULL,
+                "backup",NULL,
+                NULL);
+    exit(1);
+   }
+ else
+    HTMLMessageHead(client,200,"WWWOFFLE SSL OK",
+                    "Content-Type",NULL,
+                    NULL);
+
+ SSL_Transfer(client);
+
+ SSL_Close();
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Make a new request.
 
